# -*- coding: UTF-8 -*-
#   icd_cam2spline.py     2020.7.21      kita

#!! DXF変換されたカム曲線をモデリングする際の補助スクリプト

#   参考    icd_test_cam.py

#   必要なファイル
#   icd_module.py


import clr                  # .NETをimport
clr.AddReference('sxnet')   # 
from sxnet import *     # 全てのクラスをimport
SxSys.init(3999)
import numpy as np
import sys
import icd_module as imd

opt = SxOptSelect()
#opt.ent_mode = SxSys.ENTMODE_SEG       # 検索モード定数:0: セグメント検索   要素(寸法、直線等)の検索モードを、[要]にする
    # SxSys.ENTMODE_SEG:実像部品/パーツのメンバセグメントを選択した場合に、 セグメントを返答します。
#opt.ent_mode = SxSys.ENTMODE_PART      # 検索モード定数:1： 実像部品/パーツ検索  
opt.ent_mode = SxSys.ENTMODE_MULTI     # 検索モード定数:2: セグメントと実像部品/パーツ切換検索。
    # 実像部品、パーツのメンバセグメントを選択した場合に、CAD上でのコーディネートメニューの検索モード[部]/[要]の状態に従って、
    # セグメント又は実像部品/パーツを返答します。

# 寸法や文字を選択対象にする場合は ENTMODE_SEG。
# タップやキリ穴等、実像部品を対象にする場合は ENTMODE_PART。

opt.select_ent = True
#opt.select_pos = True
opt.select_seg = True
opt.select_go = True

# 誤差
gosa = 3.0

#------------------------------------------------------------------------------
#   Segment Listよりスプライン作成用のSxPosリストを作成。短い線分除去機能付き
#------------------------------------------------------------------------------
def seg2pnt_for_spline(seg_list, lmt):
    sx_pos_list = []
    i = 0
    i_max = len(seg_list) - 1

    for seg in seg_list:
        geom = seg.getGeom()
        pnt0 = np.array([geom.sp.x, geom.sp.y])
        if i == 0:
            sx_pos_list.append(SxPos(pnt0[0], pnt0[1], 0.0))
            pre_pnt = pnt0; i += 1
            continue
        elif i < i_max:
            ll = np.linalg.norm(pnt0 - pre_pnt)
            #print(ll, lmt)
            if  ll > lmt:    #より長い
                sx_pos_list.append(SxPos(pnt0[0], pnt0[1], 0.0))
                pre_pnt = pnt0; i += 1
                continue
            else:   #より短い
                i += 1
                continue
        else:   # i == i_max
            pnt_end = pnt0 + geom.leng * np.array([geom.vec.x, geom.vec.y])
            sx_pos_list.append(SxPos(pnt_end[0], pnt_end[1], 0.0))
            
    return sx_pos_list

#------------------------------------------------------------------------------
#   Point Listより短い線分を除去する
#------------------------------------------------------------------------------
def arange_pnt_list(pl0, lmt):
    pl = []
    i = 0
    i_max = len(pl0) - 1
    #print(i_max)    
    for pnt in pl0:
        if i == 0:
            pl.append(pnt)
            pre_pnt = pnt; i += 1
            continue
        elif i < i_max:
            ll = np.linalg.norm(pnt - pre_pnt)
            #print(ll, lmt)
            if  ll > lmt:    #より長い
                pl.append(pnt)
                pre_pnt = pnt; i += 1
                continue
            else:   #より短い
                i += 1
                continue
        elif i == i_max:
            pl.append(pnt)
            
            #ll = np.linalg.norm(pnt - pre_pnt)
            #if ll > lmt:   #より長い 
            #    pl.append(pnt)
            #else:   # より短い  
            #    #pl.remove(pre_pnt)
            #    pl.append(pnt)
    return pl

#-----------------------------------------------------
#   ２つのPont List の繋ぎ目をマージ（統合）する
#-----------------------------------------------------
def merge_pnt_list(pl0,pl1):    # ２つのPont List の繋ぎ目をマージ（統合）する
    #print('処理前長さ　%d  %d'% (len(pl0),len(pl1)))
    max0 = len(pl0)-1
    max1 = len(pl1)-1
    #print(pl0[max0-1], pl1[0]) 
    if len(pl0) == 2 and len(pl1) == 2:   # 円弧　or 直線
        print('共に円弧か直線です')
        pass
    elif len(pl0) == 2 and len(pl1) > 2:    # pl1が分解された線分のPoint List
        if   np.linalg.norm(pl0[0] - pl1[0]) < gosa   :  pl1[0]   = pl0[0]
        elif np.linalg.norm(pl0[0] - pl1[max1]) < gosa:  pl1[max1] = pl0[0]
        elif np.linalg.norm(pl0[1] - pl1[0]) < gosa   :  pl1[0]   = pl0[1]
        elif np.linalg.norm(pl0[1] - pl1[max1]) < gosa:  pl1[max1] = pl0[1]
    elif len(pl0) > 2 and len(pl1) == 2:    # pl0が分解された線分のPoint List
        #print('処理前長さ　%d  %d 誤差 = %f' % (len(pl0),len(pl1),gosa))
        if   np.linalg.norm(pl0[0] - pl1[0]) < gosa  :  pl0[0] = pl1[0]
        elif np.linalg.norm(pl0[0] - pl1[1]) < gosa  :  pl0[0] = pl1[1]
        elif np.linalg.norm(pl0[max0] - pl1[0]) < gosa:
            #print('線分長さ %f' % (np.linalg.norm(pl0[max0] - pl1[0])))
            #print(pl0[-1:], pl1[0])
            #print('pl0_end x= %f y = %f' % (pl0[-1:][0], pl0[-1:][1]))  # list index out of range
            #print('pl0_end x=%f y=%f pl1_start x=%f y=%f' % (pl0[-1:][0], pl0[-1:][1], pl1[0][0], pl1[0][1]))
            pl0[max0] = pl1[0]
            #pl0[max0-1][1] = pl1[0][1]
            #pl0[-1:] = pl1[0]
            #print(pl0[-1:])
            #pl0.pop()
            #tmp_x = pl1[0][0]; tmp_y = pl1[0][1]
            #pl0.append(np.array([tmp_x, tmp_y]))
            #pl0.append(pl1[0])
        elif np.linalg.norm(pl0[max0] - pl1[1]) < gosa:  pl0[max0] = pl1[1]
        #print('処理後長さ　%d  %d'% (len(pl0),len(pl1)))
        
        
    else :      # 両方が分解された線分のPoint List
        if   np.linalg.norm(pl0[0] - pl1[0]) < gosa   :  pl1[0]   = pl0[0]
        elif np.linalg.norm(pl0[0] - pl1[max1]) < gosa:  pl1[max1] = pl0[0]
        elif np.linalg.norm(pl0[1] - pl1[0]) < gosa   :  pl1[0]   = pl0[1]
        elif np.linalg.norm(pl0[1] - pl1[max1]) < gosa:  pl1[max1] = pl0[1]
        
    #print('処理後長さ　%d  %d'% (len(pl0),len(pl1)))
    #print(pl0[max0-1], pl1[0]) 

    return pl0, pl1



#--------------------------------------
#    Line segment をnp.arrayに変換
#--------------------------------------
def seg2np(seg_list):    # Line segment をnp.array([siten, vec])に変換
    line_list = []
    for seg in seg_list:
        geom = seg.getGeom()
        siten = np.array([geom.sp.x, geom.sp.y])
        vec = geom.leng * np.array([geom.vec.x, geom.vec.y])
        line_list.append(np.array([siten, vec]))

    return (line_list)


#************************************
#   メインルーチン
#************************************
if __name__ == '__main__':
    if len(sys.argv) < 2:   #引き数無し
        lmt = 0.5   # 短い線分除去用制限値
    elif len(sys.argv) == 2:    # 引き数１個 
        lmt = float(sys.argv[1])


    print('制限値= %f' % lmt)
    menu = 1    # menu Select
    flg_del = 1 # 分解Segmentの削除Flg

    pl_list = []    # 各区間のポイントリストのリスト       
    seg_list = []   # 選択セグメントリスト
    disasm_list = []    #分解された線分のリストのリスト
    loop = True
    while loop:
        print('\n   要素を選択してください(@GOで選択終了)')
        pick = SxSys.select(opt)     # マウスピック
        if pick.status == SxInfSelect.STATUS_GO:   #@GO
            print('\n   GOが選択されました\n')
            loop = False
            break
        elif pick.status == SxInfSelect.STATUS_ENT:  # segment 選択
            seg = pick.ent   # 選択された要素(Segment)を返す
            inf = seg.getInf()
            seg_list.append(seg)
        
            if inf.type == 12:  # その他作図要素(タイプ=SEGTYPE_OTHER_DRAW)
                print('\n   その他作図要素が選択されました(%d)' % inf.type)
                #print('\n要素が選択されました\007\n')
            elif inf.type == 5:  
                print('\n   円弧が選択されました(%d)' % inf.type )
            elif inf.type == 2:  
                print('\n   直線が選択されました(%d)' % inf.type)
            elif inf.type == 201:  
                print('\n   実像部品が選択されました(%d)' % inf.type)
            else:
                print('\n   対象外の要素です(%d)\007\n' % inf.type)
                


    for seg in seg_list:        
        inf = seg.getInf()
        #print(inf, inf.type)
        #inf_d = tmp.getInfDetail() #'SxEntSeg' object has no attribute 'getInfDetail'
        #print(inf_d.cs.org.x)
        if inf.type == 12:  # その他作図要素(タイプ=SEGTYPE_OTHER_DRAW) 
            print('\n   その他作図要素を分解します')
            line_list = seg.free()   # 分解　直線に分解される
            #print(line_list)
            pnt_list = []
            for line in line_list:
                geom = line.getGeom()
                #pnt_list.append([geom.sp.x, geom.sp.y])   # 「始点」の集まり
                pnt_list.append(np.array([geom.sp.x, geom.sp.y]))   # 「始点」の集まり
                
            endp_x = geom.sp.x + geom.leng * geom.vec.x
            endp_y = geom.sp.y + geom.leng * geom.vec.y
            #pnt_list.append([endp_x, endp_y])   # 終点追加
            pnt_list.append(np.array([endp_x, endp_y]))   # 終点追加
            disasm_list.append(line_list)  # 分解された直線のリスト
            pl_list.append(pnt_list)      # Point List のリストに追加

            if flg_del == 1:              # 分解線分の除去  
                SxEnt.delete(line_list)

        elif inf.type == 201:    # 実像部品
            print('\n   実像部品より要素を取得しています。')
            line_list = seg.getEntList()
            #print(line_list)

            
            pnt_list = []
            for line in line_list:
                geom = line.getGeom()
                print(geom)
                pnt_list.append(np.array([geom.sp.x, geom.sp.y]))   # 「始点」の集まり
                
            endp_x = geom.sp.x + geom.leng * geom.vec.x
            endp_y = geom.sp.y + geom.leng * geom.vec.y
            pnt_list.append(np.array([endp_x, endp_y]))   # 終点追加

            # 長手方向でソート
            w_x = max(pnt_list, key = lambda x:x[0])[0] - min(pnt_list, key = lambda x:x[0])[0]  # x方向の幅
            w_y = max(pnt_list, key = lambda x:x[1])[1] - min(pnt_list, key = lambda x:x[1])[1]  # y方向の幅
            #print(w_x, w_y)

            if w_x >= w_y : # x方向に長い
                pnt_list1 = sorted(pnt_list, key=lambda x: x[0])       #x座標値sort
            else:           # y方向に長い
                pnt_list1 = sorted(pnt_list, key=lambda x: x[1])       #y座標値sort
                
            disasm_list.append(pnt_list1)  # 分解された直線のリスト
            pl_list.append(pnt_list1)      # Point List のリストに追加

            if flg_del == 1:              # Segment削除  
                seg.delete()

            
            
        elif inf.type == 5:               # 円弧
            geom = seg.getGeom()
            xs = geom.cp.x + (geom.radius*np.cos(geom.sang*np.pi/180))
            ys = geom.cp.y + (geom.radius*np.sin(geom.sang*np.pi/180))
            xe = geom.cp.x + (geom.radius*np.cos((geom.sang+geom.eang)*np.pi/180))
            ye = geom.cp.y + (geom.radius*np.sin((geom.sang+geom.eang)*np.pi/180))

            pnt_s = np.array([xs, ys])
            pnt_e = np.array([xe, ye])

            #cp0 = np.array([geom.cp.x, geom.cp.y])
            #p0 = np.array([np.cos(geom.sang*np.pi/180), np.sin(geom.sang*np.pi/180)]) * geom.radius
            #p1 = np.array([np.cos((geom.sang+geom.eang)*np.pi/180), np.sin((geom.sang+geom.eang)*np.pi/180)]) * geom.radius
            #pnt_s = cp0 + p0
            #pnt_e = cp0 + p1
            #print('pnt_s  pnt_e')
            #print(pnt_s); print(pnt_e)
            pl_list.append([pnt_s, pnt_e])   # Point List のリストに追加     
            #pl_list.append(np.array([pnt_s, pnt_e]))   # Point List のリストに追加     
            #pl_list.append(np.array([cp0+p0, cp0+p1]))   # Point List のリストに追加     

            print('\n   中心座標 %f %f'% (geom.cp.x, geom.cp.y))

        elif inf.type == 2:         # 直線          
            geom = seg.getGeom()
            endp_x = geom.sp.x + geom.leng * geom.vec.x
            endp_y = geom.sp.y + geom.leng * geom.vec.y
            p0 = np.array([geom.sp.x, geom.sp.y])
            p1 = np.array([endp_x, endp_y])
            pl_list.append([p0, p1])    # Point List のリストに追加   
            #pl_list.append(np.array([p0, p1]))    # Point List のリストに追加   
                

            print('\n   始点  %f %f  終点 %f %f'% (geom.sp.x, geom.sp.y, endp_x, endp_y))
            #print('    始点  %f %f  Vec %f %f'% (geom.sp.x, geom.sp.y, geom.vec.x, geom.vec.y))
        else:
            print('\n   対象外の要素です\007\n')

    # end of for seg

    #---------------------------------------------------------
    #   ちゃんとした閉空間が構築されるよな自由曲線を作成する 本チャン
    #----------------------------------------------------------

    if menu == 1:
        print('\n   自由曲線を作成しています\n')
        #pl0_list = pl_list
        #i = 0
        i_max = len(pl_list) - 1
        #print(len(pl_list[0]), pl_list[0])
        if i_max == 1:  #要素が二つなら
            pl_list[0], pl_list[1] = merge_pnt_list(pl_list[0], pl_list[1])
            pl_list[1], pl_list[0] = merge_pnt_list(pl_list[1], pl_list[0])
        elif i_max > 1:
            for i in range(i_max+1):
                #print(pl_list[i])
                if i >= i_max:
                    pl_list[i], pl_list[0] = merge_pnt_list(pl_list[i], pl_list[0])
                else:
                    pl_list[i], pl_list[i+1] = merge_pnt_list(pl_list[i], pl_list[i+1])
                #i += 1
            
            
        for pl in pl_list:
            if len(pl) == 2:    # 直線、円弧除外
                continue
            else:   
                sx_pnt_list = []
                
                pl_arange = arange_pnt_list(pl, lmt)    # 短い線分の除去(不要?)
                
                print('制限値 %f 元の長さ %d  アレンジ後の長さ %d' % ( lmt, len(pl), len(pl_arange)))
                
                # 長手方向でソート
                w_x = max(pl_arange, key = lambda x:x[0])[0] - min(pl_arange, key = lambda x:x[0])[0]  # x方向の幅
                w_y = max(pl_arange, key = lambda x:x[1])[1] - min(pl_arange, key = lambda x:x[1])[1]  # y方向の幅
            
                #print(w_x, w_y)

                if w_x >= w_y : # x方向に長い
                    pl_arange1 = sorted(pl_arange, key=lambda x: x[0])       #x座標値sort
                else:           # y方向に長い
                    pl_arange1 = sorted(pl_arange, key=lambda x: x[1])       #y座標値sort

                for pnt in pl_arange1:
                    #print(cnt, pnt[0], pnt[1])
                    sx_pnt_list.append(SxPos(pnt[0], pnt[1], 0.0))

                seg = imd.spline2(sx_pnt_list, SxSys.COLOR_BLUE_GREEN, SxSys.STYLE_SOLID, SxSys.WIDTH_MIDDLE)
                #seg = SxEntSeg.createSpline2D(sx_pnt_list,SxEntSeg.SPLINE_KIND_NORMAL,False,False,0.0,0.0)
               
        print('\n   自由曲線の作成を終了しました\007\n')


 
    #---------------------------------------------------------
    #   分解された線分の長さを得る  Test用
    #----------------------------------------------------------

    elif menu == 2:
        len_list = []   # 長さリスト
        if len(disasm_list) != 0:
            sx_pos_list = []    # 分解された直線の始点リスト
            offset = 10
            cnt = 0; lmt = 1.0
            print('\n   分解された直線の数は ' + str(len(disasm_list[0])) + ' です')
            for ls in disasm_list[0]:     # ls : Line Segment
        
                #inf = ls.getInf()
                #print(inf)      # sxnet.SxInfEnt
                #print(inf.type) # 2
                geom = ls.getGeom()
                #print(geom.leng)
                len_list.append(geom.leng)
                if geom.leng < lmt :    cnt+=1
                tmp_pos = SxPos((geom.sp.x + offset),geom.sp.y) 
                sx_pos_list.append(tmp_pos)

            # 自由曲線（スプライン）作図
            seg = SxEntSeg.createSpline2D(sx_pos_list,SxEntSeg.SPLINE_KIND_NORMAL,False,False,0.0,0.0)
                    # 通過点座標列(最小3点～最大800点),  自然曲線, 始角、終角設定, 始点、終点接線角度(度)
            print('\n   最小長さ %f\n' % (min(len_list)))
            print('\n   %f 以下の長さの線分個数は %d です\n' % (lmt, cnt))




    #---------------------------------------------------------
    #   セグメントをアレンジせずに自由曲線を作成する  Test用
    #----------------------------------------------------------
    elif menu ==3:  # seg2pnt_for_spline()使用例

        # 長さの制限を付けて自由曲線作図
        sx_pnt_list = seg2pnt_for_spline(disasm_list[0], lmt)
        print('sx_pnt_listの個数は %d です' % (len(sx_pnt_list)))
        seg = SxEntSeg.createSpline2D(sx_pnt_list,SxEntSeg.SPLINE_KIND_NORMAL,False,False,0.0,0.0)


    

